/*
 * Copyright (c) 2022 Libre Solar Technologies GmbH
 *
 * SPDX-License-Identifier: Apache-2.0
 */
#define MODULE main

#include <zephyr/kernel.h>
#include <zephyr/logging/log.h>
#include <zephyr/device.h>
#include <zephyr/drivers/uart.h>

LOG_MODULE_REGISTER(MODULE, LOG_LEVEL_DBG);

#include <string.h>

/* change this to any other UART peripheral if desired */
#define UART_DEVICE_NODE DT_CHOSEN(nordic_nus_uart)

#define STACKSIZE 		1024
#define PRIORITY		7
#define UART_BUF_SIZE	40

static const struct device *uart = DEVICE_DT_GET(UART_DEVICE_NODE);
static struct k_work_delayable uart_work;

struct uart_data_t {
	void* 		fifo_reserved;
	uint8_t 	data[UART_BUF_SIZE];
	uint16_t 	len;
};

static K_FIFO_DEFINE(fifo_uart_tx_data);
static K_FIFO_DEFINE(fifo_uart_rx_data);

static void uart_cb(const struct device *dev, struct uart_event *evt, void* user_data)
{

	struct uart_data_t *buf;

	static size_t aborted_len;
	static uint8_t *aborted_buf;
	
	static bool disable_req;

	switch (evt->type) {
	case UART_TX_DONE:
		LOG_DBG("UART_TX_DONE");
		break;

	case UART_RX_RDY:
		LOG_DBG("UART_RX_RDY");
		buf = CONTAINER_OF(evt->data.rx.buf, struct uart_data_t, data);
		buf->len += evt->data.rx.len;
		LOG_DBG("buf->len = %d || buf->data = %s", buf->len, buf->data);
		if (disable_req) {
			LOG_WRN("DISABLE_REQ");
			return;
		}

		if(evt->data.rx.buf[buf->len-1] == '~')
		{
			LOG_WRN("~ detected");
			disable_req = true;
			uart_rx_disable(uart);
		}

		break;

	case UART_RX_DISABLED:
		LOG_DBG("UART_RX_DISABLED");
		disable_req = false;

		buf = k_malloc(sizeof(*buf));
		if (buf) {
			buf->len = 0;
		} else {
			LOG_WRN("Not able to allocate UART receive buffer");
			k_work_reschedule(&uart_work, K_MSEC(50));
			return;
		}

		uart_rx_enable(uart, buf->data, sizeof(buf->data),
			       50*1000*1000); //uS

		break;

	case UART_RX_BUF_REQUEST:
		LOG_DBG("UART_RX_BUF_REQUEST");
		buf = k_malloc(sizeof(*buf));
		if (buf) {
			buf->len = 0;
			uart_rx_buf_rsp(uart, buf->data, sizeof(buf->data));
			LOG_INF("sizeof(buf->data) = %d", sizeof(buf->data));
		} else {
			LOG_WRN("Not able to allocate UART receive buffer");
		}

		break;

	case UART_RX_BUF_RELEASED:
		LOG_DBG("UART_RX_BUF_RELEASED");
		buf = CONTAINER_OF(evt->data.rx_buf.buf, struct uart_data_t,
				   data);

		if (buf->len > 0) {
			k_fifo_put(&fifo_uart_rx_data, buf);
		} else {
			k_free(buf);
		}

		break;

	case UART_TX_ABORTED:
		LOG_DBG("UART_TX_ABORTED");
		break;

	default:
		break;
	}
}

static void uart_work_handler(struct k_work *item)
{
	struct uart_data_t *buf;

	buf = k_malloc(sizeof(*buf));
	if (buf) {
		buf->len = 0;
	} else {
		LOG_WRN("Not able to allocate UART receive buffer");
		k_work_reschedule(&uart_work, K_MSEC(50));
		return;
	}

	uart_rx_enable(uart, buf->data, sizeof(buf->data), 50000);
}

static int uart_init(void) 
{
	int err = 0;

	struct uart_data_t *rx;

	if(!device_is_ready(uart))
		return -ENODEV;
	rx = k_malloc(sizeof(*rx));
	if(rx)
		rx->len = 0;
	else
		return -ENOMEM;
	
	k_work_init_delayable(&uart_work, uart_work_handler);
	
	err = uart_callback_set(uart, uart_cb, NULL);
	if (err) {
		k_free(rx);
		LOG_ERR("No UART CB");
		return err;
	}
	
	const uint32_t RX_TIMEOUT_US = 50*1000;
	err = uart_rx_enable(uart, rx->data, sizeof(rx->data), RX_TIMEOUT_US);
	if(err) {
		LOG_ERR("Cannot enable uart reception (err: %d)", err);
		k_free(rx);
	}
	return 0;
}

int main(void)
{
	int err = 0;

	err = uart_init();
	if(err) {
		LOG_ERR("UART Init Fail (%d)", err);
		while(1) { k_msleep(1000); }
	}

	while(1) {
		k_msleep(1000);
	}

}

// empty the buffer by logging to console
void fifo_clear_fn(void)
{
	for (;;) {
		struct uart_data_t *buf = k_fifo_get(&fifo_uart_rx_data, K_FOREVER);
		
		char o_str[buf->len+1];
		memcpy(o_str, buf->data, buf->len);
		o_str[buf->len] = '\0';

		LOG_INF("%-40s", o_str);
		k_free(buf);
		// k_msleep(100);
	}
}

K_THREAD_DEFINE(
	fifo_clear_thread, 
	STACKSIZE, 
	fifo_clear_fn, 
	NULL, NULL, NULL, 
	PRIORITY, 
	0, 	// options
	0	// delay
);